"""
Parse GPS data into GeoJSON Features
"""
import geojson


class GeoJSONParser(object):
    """
    Create a GeoJSONParser object.
    
    """

    def __init__(self):
        self._collections_dict = {}

    def __iter__(self):
        """Return an iterator for `_collections_dict` items."""
        return iter(self._collections_dict.items())

    def add_feature(self, title, lat, long, properties={}):
        """
        Add a `Feature' to `_collections_dict`.

        Parameters
        ----------
        title : str
            The `FeatureCollection` title.
        lat : float
            The latitude of the Feature.
        long : float
            The longitude of the Feature.
        properties : dict
            The Feature properties.
        """
        point = geojson.Point((lat, long))
        feature = geojson.Feature(
            geometry=point, 
            properties=properties
        )
        if title not in self._collections_dict:
            feature_collection = geojson.FeatureCollection(
                features = [feature], 
                title = title
            )
            self._collections_dict[title] = feature_collection
        else:
            self._collections_dict[title]['features'].append(feature)



"""
Convert latitude / longitude DMS representation to decimal.
"""
from decimal import Decimal, getcontext
getcontext().prec = 9

SIX_PLACES = Decimal(10) **-6 
NORTH_REF = 'N'
SOUTH_REF = 'S'
EAST_REF = 'E'
WEST_REF = 'W'
MAX_MINUTES = 59
MAX_SECONDS = 59
MAX_LAT_DEGREES = 90
MAX_LONG_DEGREES = 180


def is_latitude(ref):
    """bool: Return True if `ref` is 'N' or 'S'."""
    return (ref == NORTH_REF or ref == SOUTH_REF)

def is_longitude(ref):
    """bool: Return True if `ref` is 'E' or 'W'."""
    return (ref == EAST_REF or ref == WEST_REF)

def dms_to_decimal(deg, min, sec, ref):
    """
    Convert degrees, minutes, seconds and reference to decimal.

    Parameters
    ----------
    deg : float
        Degrees.
    min : float
        Minutes.
    sec : float
        Seconds 
    ref : str
        The compass reference.
 
    Returns
    -------
    float
        Decimal degree.

    Raises
    ------
    ValueError
        If `ref` is invalid.
        If `ref` is invalid.
    """
    if ref not in [NORTH_REF, SOUTH_REF, EAST_REF, WEST_REF]:
        raise ValueError(f'ValueError: Invalid GPS Reference {ref}, Expecting N, S, E or W')
    
    if sec > MAX_SECONDS or sec < 0:
        raise ValueError(f'ValueError: Invalid Seconds {str(sec)}, Should be positive and less than 60')
    
    if min > MAX_MINUTES or min < 0:
        raise ValueError(f'ValueError: Invalid Minutes {str(min)}, Should be positive and less than 60')
    
    if deg < 0:
        raise ValueError(f'ValueError: Invalid Degrees {str(sec)}, Should be positive')
    
    if is_latitude(ref) and deg > MAX_LAT_DEGREES:
        raise ValueError(f'ValueError: Latitude {str(deg) + ref}, cannot be greater than 90 degrees')
    
    elif is_longitude(ref) and deg > MAX_LONG_DEGREES:
        raise ValueError(f'ValueError: Longitude {str(deg) + ref}, cannot be greater than 180 degrees')
    
    sign = (-1 if ref == SOUTH_REF or ref == WEST_REF else 1)
    dec_deg = (Decimal(deg) + Decimal(min)/Decimal(60) + Decimal(sec)/Decimal(3600)) * sign
    return float(dec_deg.quantize(SIX_PLACES))



"""
Tests for cli
"""

import unittest
import io
from contextlib import redirect_stdout

from im2geojson.cli import create_parser, parse_args_to_dict, main


class TestParserCreate(unittest.TestCase):

    def test_create_parser(self):
        parser = create_parser()
        self.assertIsNotNone(parser)

    def test_create_parser_prog(self):
        parser = create_parser()
        self.assertEqual('im2geojson', parser.prog)
        self.assertEqual('Geojson from image metadata', parser.description)

    
class TestParserArguments(unittest.TestCase):

    def setUp(self):
        self.parser = create_parser()

    def test_parser_short_in_path(self):
        parsed = self.parser.parse_args(['-i', 'testing/in'])
        self.assertEqual('testing/in', parsed.in_dir_path)

    def test_parser_in_path(self):
        parsed = self.parser.parse_args(['--in_dir_path', 'testing/in'])
        self.assertEqual('testing/in', parsed.in_dir_path)

    def test_parser_short_out_path(self):
        parsed = self.parser.parse_args(['-o', 'testing/out'])
        self.assertEqual('testing/out', parsed.out_dir_path)

    def test_parser_out_path(self):
        parsed = self.parser.parse_args(['--out_dir_path', 'testing/out'])
        self.assertEqual('testing/out', parsed.out_dir_path)

    def test_parser_short_save_images(self):
        parsed = self.parser.parse_args(['-s'])
        self.assertTrue(parsed.save_images)

    def test_parser_save_images(self):
        parsed = self.parser.parse_args(['--save_images'])
        self.assertTrue(parsed.save_images)

    def test_parser_short_save_thumbnails(self):
        parsed = self.parser.parse_args(['-t'])
        self.assertTrue(parsed.save_thumbnails)

    def test_parser_save_thumbnails(self):
        parsed = self.parser.parse_args(['--save_thumbnails'])
        self.assertTrue(parsed.save_thumbnails)

    def test_parser_defaults(self):
        parsed = self.parser.parse_args([])
        with self.assertRaises(AttributeError):
            parsed.in_dir_path
        with self.assertRaises(AttributeError):
            parsed.out_dir_path
        with self.assertRaises(AttributeError):
            parsed.save_images
        with self.assertRaises(AttributeError):
            parsed.save_thumbnails


class TestParseArgs(unittest.TestCase):

    def test_parsed_args_dict(self):
        args = ['-i', 'testing/in', '-o', 'testing/out', '-s', '-t']
        parsed_args_dict = parse_args_to_dict(args)
        expected = {'in_dir_path': 'testing/in', 
                    'out_dir_path': 'testing/out', 
                    'save_images': True, 
                    'save_thumbnails': True}
        self.assertEqual(expected, parsed_args_dict)


class TestMain(unittest.TestCase):

    def setUp(self):
        self.out_path = 'tests/test_out_path/'
    
    def test_main(self):
        f = io.StringIO()
        with redirect_stdout(f):
            
            main(['-o', self.out_path])
        out = f.getvalue()
        out_lines = out.split('\n')

        expected_first_line = 'Running...'
        expected_last_line = '0 out of 0 images completed successfully'

        self.assertEqual(expected_first_line, out_lines[0])
        self.assertEqual(expected_last_line, out_lines[2])


if __name__ == '__main__':  
    unittest.main()             # pragma: no cover


"""
im2geojson.py
"""
import os
import glob
import json
import concurrent.futures

from .geojson_parser import GeoJSONParser
from .exif_reader import read_exif
from .timer import Timer

DEFAULT_IN_DIR_PATH = './'
DEFAULT_OUT_DIR_PATH = './'
OUT_DIR = 'im2geojson_output/'
GEOJSON_DIR = 'geojson/'
IMAGE_DIR = 'images/'


class ImageToGeoJSON(object):
    """
    ImageToGeoJSON

    Attributes
    ----------
    Notes
    -----
    Saves the harvested metadata as geojson to 'out_dir_path`
    Optionally saves images without metadata and thumbnails.
    """

    def __init__(self, 
                 in_dir_path=DEFAULT_IN_DIR_PATH, 
                 out_dir_path=DEFAULT_OUT_DIR_PATH, 
                 save_images=False, 
                 save_thumbnails=False):
        
        self._in_dir_path = in_dir_path
        self._out_dir_path = out_dir_path
        self._save_images = save_images
        self._save_thumbnails = save_thumbnails

        self._geojson_parser = GeoJSONParser()
        self._timer = None
        self._errors = {}
        self._total_count = 0
        self._success_count = 0

        # Make Output Directories
        dir_paths = [self.geojson_dir_path]
        if save_images or save_thumbnails:
            dir_paths.append(self.image_dir_path)
        for path in dir_paths:
            try:
                os.makedirs(path)
            except FileExistsError:
                # print(f"Folder {path} already exists")
                pass
            else:
                # print(f"Folder {path} created!")
                pass

    @property
    def in_dir_path(self):
        """str: Return the path to the input directory."""
        return self._in_dir_path
    
    @property
    def out_dir_path(self):
        """str: Return the path to the output directory."""
        return self._out_dir_path
    
    @property
    def geojson_dir_path(self):
        """str: Return the path to the geojson directory."""
        return os.path.join(self.out_dir_path, OUT_DIR, GEOJSON_DIR)
    
    @property
    def image_dir_path(self):
        """str: Return the path to the image directory."""
        return os.path.join(self.out_dir_path, OUT_DIR, IMAGE_DIR)

    @property
    def errors(self):
        """dict: Return the error dictionary or 'No errors'."""
        if self._errors == {}:
            return 'No errors'
        else:
            return self._errors
        
    @property
    def summary(self):
        """str: Return the summary string."""
        return f'{self._success_count} out of {self._total_count} images completed successfully'

    def start(self):
        """
        Read and process the images from `in_dir_path`.

        """
        if self._timer is not None:
            raise RuntimeError('Error: Too many calls to function')
        
        with Timer() as self._timer:
            self._process_files()
        print(self.summary)
            
    def _process_files(self):
        # Process image files concurrently
        files = glob.iglob(f'{self.in_dir_path}**/*.[Jj][Pp][Gg]')
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future_to_path = {executor.submit(self._process_image_file, filepath): filepath for filepath in files}
            for future in concurrent.futures.as_completed(future_to_path):
                filepath = future_to_path[future]
                self._total_count += 1
                try:
                    folder, coord, props = future.result()
                except Exception as e:
                    self._add_file_to_errors_with_exception_string(filepath, str(e))
                else:
                    self._geojson_parser.add_feature(folder, *coord, props)
                    self._success_count += 1

        # Save geojson
        for title, feature_collection in self._geojson_parser:
            geojson_file_path = os.path.join(self.geojson_dir_path, f'{title}.geojson')
            with open(geojson_file_path, 'w') as f:
                json.dump(feature_collection, f)

    def _process_image_file(self, filepath):
        try:
            coord, props, image_b, thumb_b = read_exif(filepath, 
                                                       get_image=self._save_images, 
                                                       get_thumbnail=self._save_thumbnails)
        except FileNotFoundError as e:
            raise FileNotFoundError('FileNotFoundError: No such file or directory') from e
        except Exception as e:
            raise e
        else:
            props['original_image_absolute_path'] = filepath
            folder, filename = ImageToGeoJSON.folder_and_filename_from_filepath(filepath)

            # image 
            if self._save_images and image_b is not None:
                rel_image_path = self._rel_image_path(filename)
                image_path = os.path.join(self.out_dir_path, rel_image_path)            

                with open(image_path, 'wb') as im:
                    im.write(image_b)
                    props["rel_image_path"] = rel_image_path

            # thumbnail 
            if self._save_thumbnails and thumb_b is not None:
                rel_thumbnail_path = self._rel_thumbnail_path(filename)
                thumbnail_path = os.path.join(self.out_dir_path, rel_thumbnail_path)

                with open(thumbnail_path, 'wb') as im:
                    im.write(thumb_b)
                    props["rel_thumbnail_path"] = rel_thumbnail_path

            return folder, coord, props
        
    def _add_file_to_errors_with_exception_string(self, filepath, exception_string):
        folder, filename = ImageToGeoJSON.folder_and_filename_from_filepath(filepath)
        key = os.path.join(folder, filename)
        self._errors[key] = exception_string

    def _rel_image_path(self, filename):
        # Return the relative path to the image filename.
        return os.path.join(OUT_DIR, IMAGE_DIR, filename)
    
    def _rel_thumbnail_path(self, filename):
        # Return the relative path to the thumbnail image filename.
        thumb_file_name = ImageToGeoJSON.thumbnail_filename_from_image_filename(filename)
        return os.path.join(OUT_DIR, IMAGE_DIR, thumb_file_name)

    @staticmethod
    def folder_and_filename_from_filepath(filepath):
        """tuple of str: Split the filepath and return the folder and filename."""
        head, filename = os.path.split(filepath)
        head, folder = os.path.split(head)
        return folder, filename
    
    @staticmethod
    def thumbnail_filename_from_image_filename(filename):
        """str: Split the image filename and return the thumbnail filename."""
        f_name, f_type  = filename.split('.')
        return f_name + '_thumb.' + f_type


"""
Tests for dms_conversion
"""

import unittest
from im2geojson.dms_conversion import dms_to_decimal


class TestDMSToDecimal(unittest.TestCase):

    def test_convert_zero_east(self):
        test_result = 0
        self.assertEqual(test_result, dms_to_decimal(0, 0, 0, 'E'))

    def test_convert_five_degrees_east(self):
        test_result = 5
        self.assertEqual(test_result, dms_to_decimal(5, 0, 0, 'E'))

    def test_convert_five_degrees_twelve_minutes_north(self):
        test_result = 5.2
        self.assertEqual(test_result, dms_to_decimal(5, 12, 0, 'N'))

    def test_convert_five_degrees_twelve_minutes_eighteen_seconds_north(self):
        test_result = 5.205
        self.assertEqual(test_result, dms_to_decimal(5, 12, 18, 'N'))

    def test_convert_five_degrees_twelve_minutes_eighteen_seconds_south(self):
        test_result = -5.205
        self.assertEqual(test_result, dms_to_decimal(5, 12, 18, 'S'))

    def test_convert_five_degrees_twelve_minutes_eighteen_seconds_west(self):
        test_result = -5.205
        self.assertEqual(test_result, dms_to_decimal(5, 12, 18, 'W'))

    def test_convert_fiftynine_degrees_five_minutes_three_seconds_north(self):
        test_result = 59.084167
        self.assertEqual(test_result, dms_to_decimal(59, 5, 3, 'N'))

    def test_lat_91N_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(91, 0, 0, 'N')

    def test_lat_91S_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(91, 0, 0, 'S')

    def test_long_181W_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(181, 0, 0, 'W')

    def test_long_181E_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(181, 0, 0, 'E')

    def test_invalid_ref_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(1, 1, 1, 'Z')

    def test_invalid_minutes_60_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(1, 60, 1, 'N')

    def test_invalid_negative_minutes_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(1, -1, 1, 'N')

    def test_invalid_seconds_60_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(1, 1, 60, 'N')

    def test_invalid_negative_seconds_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(1, 1, -1, 'N')

    def test_invalid_negative_degrees_raises_exception(self):
        with self.assertRaises(ValueError):
            dms_to_decimal(-1, 1, 1, 'N')


if __name__ == '__main__':
    unittest.main()             # pragma: no cover


"""
cli.py
"""

import argparse

from im2geojson.im2geojson import ImageToGeoJSON


def create_parser():
    parser = argparse.ArgumentParser(
        argument_default=argparse.SUPPRESS,
        prog='im2geojson',
        description='Geojson from image metadata',
        # epilog='Text at the bottom of help'
        )
    parser.add_argument(
        '-i', 
        '--in_dir_path', 
        help='The path to the images to process', 
        type=str
        )
    parser.add_argument(
        '-o', 
        '--out_dir_path', 
        help='The path to output', 
        type=str
        )
    parser.add_argument(
        '-s', 
        '--save_images', 
        help='Save Images stripped of metadata', 
        action='store_true'
        )
    parser.add_argument(
        '-t', 
        '--save_thumbnails', 
        help='Save thumbnail images', 
        action=argparse.BooleanOptionalAction
        )
    return parser

def parse_args_to_dict(args):
    parser = create_parser()
    parsed_args = parser.parse_args(args)
    parsed_args_dict = vars(parsed_args)
    return parsed_args_dict

def main(args=None):
    parsed_args_dict = parse_args_to_dict(args)
    im2geo = ImageToGeoJSON(**parsed_args_dict)
    im2geo.start()


if __name__ == '__main__':
    main(args=None)                 # pragma: no cover


"""
Read Exif metadata from image
"""
from exif import Image
from datetime import datetime
import warnings 
import threading

from .dms_conversion import dms_to_decimal


def read_exif(filepath, get_image=False, get_thumbnail=False):
    """
    Read exif metadata from image file at `filepath`.
    
    Parameters
    ----------
    filepath : str
        The path to the image file.

    Returns
    -------
    (lat, long) : tuple of float
        The decimal latitude, longitude coordinate as a float.
    props : dictionary
        Dictionary containing the date the image was captured.
    image_b : bytes
        The image stripped of exif metadata.
    thumb_b : bytes
        The thumbnail.
    
    Raises
    ------
    KeyError
        If `image_file` has no metadata.
    AttributeError
        If `image_file` has missing metadata.
    ValueError
        If `image_file` has invalid metadata.
    """
    with open(filepath, 'rb') as image_file:
        image = Image(image_file)
        if not image.has_exif:
            raise KeyError('No metadata.')

        # coord
        try:
            dms_lat = (*image.gps_latitude, image.gps_latitude_ref)
            dms_long = (*image.gps_longitude, image.gps_longitude_ref)
        except AttributeError as e:
            raise AttributeError(f'AttributeError: {e}') from e
        else:
            try:
                lat = dms_to_decimal(*dms_lat)
                long = dms_to_decimal(*dms_long)
            except ValueError as e:
                raise e
        
        # datetime
        try:
            datetime_str = image.datetime_original
        except AttributeError as e:
            raise AttributeError(f'AttributeError: {e}') from e
        else:
            try:
                datetime_object = datetime.strptime(datetime_str, '%Y:%m:%d %H:%M:%S')
            except ValueError as e:
                raise ValueError(f'ValueError: {e}') from e

        # props 
        props = { 
            "datetime": str(datetime_object),
            }

        # delete exif data
        if get_image:
            with threading.RLock():
                # Catch warning that not all data has been deleted:
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    image.delete_all()
            
        # files
        image_b = image.get_file() if get_image else None
        thumb_b = image.get_thumbnail() if get_thumbnail else None


        return (lat, long), props, image_b, thumb_b


"""
Tests for geojsonphoto
"""

import unittest
import os
import shutil
import io
from contextlib import redirect_stdout

from im2geojson.im2geojson import *


class TestImageToGeoJSONInit(unittest.TestCase):

    def setUp(self):
        self.default_in_path = "./"
        self.in_path = 'tests/test_files/test_images/test_exif/'
        self.out_path = 'tests/test_out_path/'
        self.geojson_dir_path = os.path.join(self.out_path, OUT_DIR, GEOJSON_DIR)
        self.image_dir_path = os.path.join(self.out_path, OUT_DIR, IMAGE_DIR)

    def tearDown(self):
        out_path = os.path.join(self.out_path, OUT_DIR)
        if os.path.isdir(out_path):
            shutil.rmtree(out_path)

        default_path = os.path.join(DEFAULT_OUT_DIR_PATH, OUT_DIR)
        if os.path.isdir(default_path):
            shutil.rmtree(default_path)
    
    def test_in_path(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path)
        self.assertEqual(self.in_path, im2geojson.in_dir_path)

    def test_default_in_path(self):
        im2geojson = ImageToGeoJSON()
        self.assertEqual(self.default_in_path, im2geojson.in_dir_path)
    
    def test_out_path(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, out_dir_path = self.out_path)
        self.assertEqual(self.out_path, im2geojson.out_dir_path)

    def test_init_creates_geojsonphoto_dir_path(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, out_dir_path = self.out_path)
        self.assertEqual(self.geojson_dir_path, im2geojson.geojson_dir_path)

    def test_init_creates_image_dir_path(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, out_dir_path = self.out_path)
        self.assertEqual(self.image_dir_path, im2geojson.image_dir_path)
    
    def test_init_geojsonphoto_parser(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, out_dir_path = self.out_path)
        self.assertTrue(im2geojson._geojson_parser)
    
    def test_default_init_im2geojson_creates_out_directories(self):
        self.assertFalse(os.path.isdir(os.path.join(self.out_path, OUT_DIR)))

        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                            out_dir_path = self.out_path)

        self.assertTrue(os.path.isdir(self.geojson_dir_path))
        self.assertFalse(os.path.isdir(self.image_dir_path))
    
    def test_thumbnail_init_im2geojson_creates_out_directories(self):
        self.assertFalse(os.path.isdir(os.path.join(self.out_path, OUT_DIR)))

        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                            out_dir_path = self.out_path, 
                            save_images=False, 
                            save_thumbnails=True)

        self.assertTrue(os.path.isdir(self.geojson_dir_path))
        self.assertTrue(os.path.isdir(self.image_dir_path))
    
    def test_image_init_im2geojson_creates_out_directories(self):
        self.assertFalse(os.path.isdir(os.path.join(self.out_path, OUT_DIR)))

        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                            out_dir_path = self.out_path, 
                            save_images=True, 
                            save_thumbnails=False)

        self.assertTrue(os.path.isdir(self.geojson_dir_path))
        self.assertTrue(os.path.isdir(self.image_dir_path))
        

class TestImageToGeoJSONStart(TestImageToGeoJSONInit):

    def setUp(self):
        super().setUp()
        self.test_geojson_file_name = 'test_folder.geojson'

    def test_im2geojson_start_creates_geojsonphoto_file(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                            out_dir_path = self.out_path, 
                            save_images=False)
        im2geojson.start()

        self.assertTrue(os.path.isdir(self.geojson_dir_path))

        geojson_path = os.path.join(self.geojson_dir_path, self.test_geojson_file_name)
        with open(geojson_path, 'r') as f:
            jsn = json.load(f)
            self.assertIsNotNone(jsn['features'][0]['properties']['datetime'])

            with self.assertRaises(KeyError):
                jsn['features'][0]['properties']['image_path']
            with self.assertRaises(KeyError):
                jsn['features'][0]['properties']['thumbnail_path']

    def test_im2geojson_start_creates_image_file(self):
        test_image_file = 'EXIF.jpg'
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                            out_dir_path = self.out_path, 
                            save_images=True)
        im2geojson.start()

        self.assertTrue(os.path.isdir(self.image_dir_path))

        image_path = os.path.join(self.image_dir_path, test_image_file)
        with open(image_path, 'r') as f:
            pass
        
        geojson_path = os.path.join(self.geojson_dir_path, self.test_geojson_file_name)
        with open(geojson_path, 'r') as f:
            jsn = json.load(f)
            self.assertIsNotNone(jsn['features'][0]['properties']['datetime'])
            self.assertIsNotNone(jsn['features'][0]['properties']['rel_image_path'])

            with self.assertRaises(KeyError):
                jsn['features'][0]['properties']['rel_thumbnail_path']

    def test_im2geojson_start_creates_thumbnail_file(self):
        test_thumbnail_file = 'EXIF_thumb.jpg'
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                            out_dir_path = self.out_path, 
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()

        self.assertTrue(os.path.isdir(self.image_dir_path))

        thumb_path = os.path.join(self.image_dir_path, test_thumbnail_file)
        with open(thumb_path, 'r') as f:
            pass
        
        geojson_path = os.path.join(self.geojson_dir_path, self.test_geojson_file_name)
        with open(geojson_path, 'r') as f:
            jsn = json.load(f)
            self.assertIsNotNone(jsn['features'][0]['properties']['datetime'])
            self.assertIsNotNone(jsn['features'][0]['properties']['rel_thumbnail_path'])

            with self.assertRaises(KeyError):
                jsn['features'][0]['properties']['rel_image_path']

    def test_geojsonphoto_start_original_image_absolute_path(self):
        test_thumbnail_file = 'EXIF_thumb.jpg'
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                            out_dir_path = self.out_path, 
                            save_images=False, 
                            save_thumbnails=False)
        im2geojson.start()

        geojson_path = os.path.join(self.geojson_dir_path, self.test_geojson_file_name)
        with open(geojson_path, 'r') as f:
            jsn = json.load(f)
            self.assertIsNotNone(jsn['features'][0]['properties']['original_image_absolute_path'])


class TestImageToGeoJSONStatus(TestImageToGeoJSONInit):

    def test_in_progress_status(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                                 out_dir_path = self.out_path, 
                                 save_images=False, 
                                 save_thumbnails=True)
        f = io.StringIO()
        with redirect_stdout(f):
            im2geojson.start()
        out = f.getvalue()
        self.assertIn('Running...\nFinished in ', out)

    def test_repeat_calls_to_start_raises_exception(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                                 out_dir_path = self.out_path, 
                                 save_images=False, 
                                 save_thumbnails=True)
        im2geojson.start()
        with self.assertRaises(RuntimeError):
            im2geojson.start()


class TestImageToGeoJSONErrors(TestImageToGeoJSONInit):
    
    def test_no_errors(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                            out_dir_path = self.out_path, 
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()
        self.assertEqual('No errors', im2geojson.errors)
    
    def test_errors(self):
        in_path = 'tests/test_files/test_images/test_no_exif/'
        im2geojson = ImageToGeoJSON(in_dir_path = in_path, 
                            out_dir_path = self.out_path,
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()
        test_folder_file = 'test_folder/NO_EXIF.jpg'
        test_error_dictionary = {
            test_folder_file:"'No metadata.'"
        }
        self.assertEqual(test_error_dictionary, im2geojson.errors)


class TestImageToGeoJSONSummary(TestImageToGeoJSONInit):

    def test_summary(self):
        im2geojson = ImageToGeoJSON(in_dir_path = self.in_path, 
                            out_dir_path = self.out_path, 
                            save_images=False, 
                            save_thumbnails=True)
        im2geojson.start()
        self.assertEqual('1 out of 1 images completed successfully', im2geojson.summary)
        
        
class TestFolderFilesFromPath(unittest.TestCase):

    def setUp(self):
        self.test_filename = 'image_file.jpg'
        self.test_folder_filename = 'image_file_thumb.jpg'
        self.test_folder_name = 'folder'
        self.test_in_path = os.path.join('tests/test_files/', self.test_folder_name, self.test_filename)

    def test_filename_from_path(self):
        filename = ImageToGeoJSON.folder_and_filename_from_filepath(self.test_in_path)[1]
        self.assertEqual(self.test_filename, filename )

    def test_folder_from_path(self):
        folder = ImageToGeoJSON.folder_and_filename_from_filepath(self.test_in_path)[0]
        self.assertEqual(self.test_folder_name, folder)

    def test_thumbnail_filename_from_image_filename(self):
        thumbnail_filename = ImageToGeoJSON.thumbnail_filename_from_image_filename(self.test_filename)
        self.assertEqual(self.test_folder_filename, thumbnail_filename)



if __name__ == '__main__':
    unittest.main()             # pragma: no cover